/* C8051F300 Rx code for CH-12 433mhz transceiver
   Jean Louis Coural 10/02/2021  GNU licence
	 https://github.com/merbanan/C8051F300_examples/tree/master/uart
	 https://technobyte.org/8051-uart-serial-communication-working/
	 
	 tx/rx numeric 
	 protocol: 
		asynchrone master slave UART
		1 start bit 8 data bits 1 stop bit
		baud rate 9600
	 frame 6 bytes
		StartByte Rudder Motor Activation CRC EndByte
		frame rate: on Tx PPM = 22.5 ms
		
		protection bit:
		:011DFF00F1F2  (set it before :00000001FF  in .HEX file)
	 
*/


#include <C8051F300.h>
#include <stdio.h>
#include <hc12.h>



void main (void)
{	
	Init_Device();
	//wait(20);
	INIT_TORP(TORP1);
	
	while(1)
	{		
		if(NewFrame)
		{			
			unsigned char i;
			for (i = 0; i < DATASIZE; i++) tmp_arr[i] = uart_rx_buf[i];
			NewFrame = 0; 
			if( CRC8(tmp_arr,3) == tmp_arr[CRC] )
			{	
				update_data(ACTIV);
				if(active == ID) update_data(MOT);				
				if(active == ID) update_data(SERV);								
			}						
		}	
	}
}


void INIT_TORP(unsigned char id){
	ID = id;						// specific for each torpedo	
	active = 0;		
	NewFrame = 0;
	STOP();		
}

/*-------------------------------------------------------------------
PCA_isr triggered on PCA0 counter overflow
	ECF = 1   (PCA0MD.0)
	EPCA0 = 1 (EIE1.4)
	8 khz (SYSCLCK/12)  T=125uS  2T= 1 RTC unit   Tpca=0.488 us
	24 khz  SYSCLK/4    Tpwm= 41.795 uS = Tpca*256	Tpca = 0.16326 us
	temps max (FFFF) = 10.7 ms
--------------------------------------------------------------------*/

void PCA_isr(void) interrupt 9 
{
	if(CF) {			
			//	PCA0L = 0x63;		 // 0x63  13 Khz  SYSCLK/12
			//	PCA0L = 0x8E;		 // 0x8E  16 Khz
			//	PCA0L = 0x98;		 // 0x98  20 Khz	
						PCA0L = 0x55;	 // 0x56  12 Khz   <---
						PCA0H = 0xFF;						
							
						PCA0CPH0  = DCM;		// Update motor duty cycle
						PCA0CPH1  = DCS;		// Update servo duty cycle
						
						CF = 0;					// clear overflow flag PCA0CN.7											
				}					

}	

void UART0_ISR (void) interrupt 4
{
	unsigned char dat = 0;								
	
	if(RI0)					// if received byte completed
	{
		RI0 = 0;			// Clear interrupt
		dat = SBUF0;	// Add received byte to buffer
		tx_busy = 1;	 
		
		if(NewFrame) return;

		if (dat == StartByte)
		{							    
			BYTEinCount = 0;		
			//for (unsigned char i = 0; i < DATASIZE; i++) uart_rx_buf[i] = 0;
			return;
		}
			
		if (dat == EndByte)
		{
			NewFrame = 1;
			return;
		}

		if(BYTEinCount < DATASIZE)
		{ 
			uart_rx_buf[BYTEinCount] = dat;
			BYTEinCount++;
		}
	
	}
		
	
	if (TI0) {					// if transmit byte completed
		TI0 = 0;					// Clear interrupt
		tx_busy = 0;			// UART ready to transmit next byte
	}
}


/*-------------------------------------------------------------------
 WAIT routine      
 200	= 1271us
 1 		= 6.355 us
---------------------------------------------------------------------*/

void wait(unsigned int t)
{ unsigned char i;
  unsigned int j;

  for(i=0; i<10; i++)
    { for(j=0;j<t;j++);}

}


void update_data(unsigned char target) 
{	
	unsigned char old;
	
	old = sdir;	
	
switch(target) 
	{
		
		case ACTIV:
			if(active == ID && tmp_arr[ACTIV] != ID) STOP();
			if(!active && tmp_arr[ACTIV] == ID) START();
			active = tmp_arr[ACTIV];
			break;
		
		case SERV:	
			if(tmp_arr[SERV] == 0) break;
			if(tmp_arr[SERV] > PULSMID)
			{
				tmp = tmp_arr[SERV] - PULSMID;   // {170,255} --> {0,85}
				sdir = RIGHT;
			}
			else
			{
				tmp = PULSMID - tmp_arr[SERV];	// {170,85} --> {0,85}
				sdir = LEFT;					
			}	
			
			if(tmp < DZ) { DCS = 0xFF; sdir = NEUTRAL;}
			else 
				{ if(tmp >= 85)  DCS = 0x55;
					else
					{	tmp = tmp << 1;						// {0,85} --> {0, 170}
						DCS = (unsigned char) tmp;
						DCS = ~DCS;						    // {0, 170}  --> {0xFF, 0x55}
					}
				}	
				
			if(sdir != old) reconfig_servo();
			break;
			
		case MOT:
			if(tmp_arr[MOT] == 0) break;
			DCM = tmp_arr[MOT];
			if(DCM > 250) DCM = 255;
			break;
			
		default: 
			break;
	}
				
}								
											
																						

// enable PWM on activation
void START(void) {
	XBR2 &= ~0x40; 				// disable crossbar
	PCA0CPH0  = 0xFF;			// set duty cycle to stall
	PCA0CPH1  = 0xFF;	
	PCA0CPM0  = 0x42; 		// PWM 0N
	PCA0CPM1  = 0x42; 
	//XBR0 = 0x0C;				// enable motor + servo right
	//XBR0 = 0x14;				// enable motor + servo left
	XBR0 = 0xCD;					// enable motor only	
	DCM = 0xFF;
	DCS = 0xFF;
	sdir = NEUTRAL; 	
	XBR2 |= 0x40;         // enable crossbar 
}



// enable PWM off desactivation
void STOP(void) {
	//IE=0 ?
	XBR2 &= ~0x40; 			// disable crossbar
	PCA0CPH0  = 0xFF;		// set duty cycle to stall
	PCA0CPH1  = 0xFF;	
	PCA0CPM0  = 0x00; 	// PWM off
	PCA0CPM1  = 0x00;
	XBR0 = 0xCF; 				// skip all except uart
	P0 &= ~0x0E;	
	DCM = 0xFF;
	DCS = 0xFF;
	sdir = NEUTRAL;
	XBR2 |= 0x40;       // enable crossbar 
}

/*
G1 P02  
G2 P03
right P02 pwm  P03 low
left  P03 low  P02 pwm

SKIP patterns (4&5 = uart never skipped)
  54 321
1100 1111  0xCF all skipped all off
1100 1001  0xC9 skip for motor + servo right (cex0 pin1 cex1 pin 2)
1100 0101  0xC5 skip for motor + servo left  (cex0 pin1 cex1 pin 3)
1100 1101  0xCD skip for motor alone  (cex0 pin1 cex1 off)
*/
void reconfig_servo(void){
	
	XBR2 &= ~0x40; 				// disable crossbar
	
	if(sdir == RIGHT ) // turn right
	{
		XBR0 = 0xC9;			// enable CEX0 (motor) on P01 and CEX1 on P0.2 (servo right)
		//G1 = 1;
		//G2 = 0;
		PCA0CPM1  = 0x42;
	}
	
	if(sdir == LEFT) // turn left
	{
		XBR0 = 0xC5;			// enable CEX0 (motor) on P01 and CEX1 on P0.3 (servo left)
		//G1 = 0;
		//G2 = 1;
		PCA0CPM1  = 0x42;
	}
	
	if(sdir == NEUTRAL)  // neutral
	{
		XBR0 = 0xCD;	// enable CEX0 (motor) on P01 only
		G1=0;
		G2=0;
		PCA0CPM1  = 0x00;		
	}
	
	XBR2 |= 0x40;      // enable crossbar  
}


unsigned char CRC8(unsigned char *addr, unsigned char len)
{
	unsigned char crc = 0, i=0, j=0;
	
	for (i=0; i<len;i++)
		{
			unsigned char inbyte = addr[i];
			for (j=0;j<8;j++)
			{
				unsigned char mix = (crc ^ inbyte) & 0x01;
				crc >>= 1;
				if (mix)
					crc ^= CRC_POLY;

				inbyte >>= 1;
			}
		}

	return crc;	
}

/*
void reset_timer2(void){
	TMR2L	  = 0x00;	
	TMR2H	  = 0x00; 
}


void send_byte(unsigned char byte)
{
	while(tx_busy);			// Wait until tx queue is free
	tx_busy = 1;			// Set uart tx busy
	SBUF0=byte;				// Transmit byte
}

void send_string(unsigned char *string)
{
	while(*string)
		send_byte(*string++);	// Send string 1 byte at a time
}

*/
/*-------------------------------------------------------------------
 Microcontroler initialisations
---------------------------------------------------------------------*/

void Reset_Sources_Init()
{    
   //VDM0CN    = 0x80;	// Vdd monitor ON
   wait(20);  					// Wait > 100us for initialization
   RSTSRC    = 0x02;		// Vdd monitor enabled
   //RSTSRC = 0x04;				// Enable missing clock detector
   wait(20); 
}
	
void Oscillator_Init()
{
    OSCICN    = 0x17;  // Enable internal oscillator (24.5 MHz)
}



void Voltage_Reference_Init()
{
    REF0CN    = 0x08;  // REF=Vdd sinon adc marche pas
}

void PCA_Init()
{
    PCA0CN    = 0x40;   // pca counter enabled
    PCA0MD    &= ~0x40;	// WDT OFF
		PCA0MD    = 0x01;	  // PWM interrupt on overflow  SYSCLK/12  1 tick=0.49 uS
												
    PCA0CPM0  = 0x00;   // 0x00 PWM OFF     (0x4B PWM ON/ CC isr enabled)
												// 0x42 PWM ON/ CC isr disabled
		PCA0CPM1  = 0x00;

		PCA0H = 0xFF;
		PCA0L = 0x00;	  		
}

void UART_Init()
{
// Configure UART for 9600 baud, 8 data bits, 1 stop bit, no parity
	TMOD = 0x20;			// T1 mode 2 8-bit counter/timer with autoreload
	CKCON = 0x00;  			// T1 SYCLK/12	
	TH1 = 0x96;    			// 9600 bauds
	SCON0 = 0x10;   		// Rx mode enabled	
	TR1 = 1;				// Enable Timer 1	
	
}

void Port_IO_Init()
{
	// P0.0   N/C    	digital open drain	
	// P0.1   Motor  	digital push pull
	// P0.2   S1a 	 	digital push pull
	// P0.3   S1b 		digital push pull
	// P0.4   Tx 		digital push pull
	// P0.5   Rx 		digital open drain
	// P0.6	  N/C		digital open drain 
	
	P0MDIN    = 0xFF;   // all digital	
	P0MDOUT   = 0x1E;   // all open drain exept P0.1 P0.2 P0.3 & P0.4 (push pull)	
	XBR0 	  = 0xCF;	// UART is always on P0.4 and P0.5(not skipped)	
	
	XBR1      = 0x83;	// CEX0 & CEX1 routed at port pin; Enable UART RX and TX
//	XBR2      = 0x00;	// disable crossbar & enable weak pullups
	XBR2      = 0x80;	// disable crossbar & disable weak pullups
		  
}


/* timers
	  SYCCLK = 1/24.5 mhz = 0.0408 uS
	  prescaler  1:48  1 tick = 1.959 uS
	  SYSCLK/12  1 tick = 0.49 us		
*/
/*
void Timer_Init()
{
	

	TMR2CN    = 0x04;	// timer2 16bit  SYSCLK/12  1 tick = 0.48979 us
	TMR2L	  = 0x00;	
	TMR2H	  = 0x00;  	
	
}
*/

void Interrupts_Init()
{
//---UART	
	IP = 0x10;		// Set UART interrupts to high priority
	ES0 = 1;		  // Enable UART0 interrupts	
//---PCA
	//EIE1 &= ~0x08;  // disable PCA interrupt
	EIE1  |= 0x08;    // enable PCA interrupt
	//EIP1  |= 0x08;	// pca high priority
//---GIE	
	EA = 1;  				  // enable interrupts
	//IE        = 0xA0;		  // INT  ON  avec T0 isr OFF   T1 isr OFF  T2 isr ON
	
}

void Init_Device(void)
{
	unsigned char i;
	for (i = 0; i < DATASIZE; i++) uart_rx_buf[i] = 0;
	
	PCA0MD &= ~0x40;			// Disable the watchdog timer	
	Oscillator_Init();	
	Reset_Sources_Init();
	Voltage_Reference_Init();
	Port_IO_Init();
	UART_Init();	
	PCA_Init();
	//ADC_Init();
	//Timer_Init();
	Interrupts_Init();  
					
}